@@ -64,12 +64,11 @@ typedef enum
    BSON_JSON_LF_DATE,
    BSON_JSON_LF_TIMESTAMP_T,
    BSON_JSON_LF_TIMESTAMP_I,
-   BSON_JSON_LF_REF,
-   BSON_JSON_LF_ID,
    BSON_JSON_LF_UNDEFINED,
    BSON_JSON_LF_MINKEY,
    BSON_JSON_LF_MAXKEY,
    BSON_JSON_LF_INT64,
+   BSON_JSON_LF_DECIMAL128
 } bson_json_read_bson_state_t;
 
 
@@ -114,11 +113,6 @@ typedef union
       uint32_t t;
       uint32_t i;
    } timestamp;
-   struct {
-      bool       has_ref;
-      bool       has_id;
-      bson_oid_t id;
-   } ref;
    struct {
       bool has_undefined;
    } undefined;
@@ -131,6 +125,11 @@ typedef union
    struct {
       int64_t value;
    } v_int64;
+#ifdef BSON_EXPERIMENTAL_FEATURES
+   struct {
+      bson_decimal128_t value;
+   } v_decimal128;
+#endif
 } bson_json_bson_data_t;
 
 
@@ -460,10 +459,9 @@ _bson_json_read_integer (void    *_ctx, /* IN */
       case BSON_JSON_LF_OID:
       case BSON_JSON_LF_BINARY:
       case BSON_JSON_LF_TYPE:
-      case BSON_JSON_LF_REF:
-      case BSON_JSON_LF_ID:
       case BSON_JSON_LF_UNDEFINED:
       case BSON_JSON_LF_INT64:
+      case BSON_JSON_LF_DECIMAL128:
       default:
          _bson_json_read_set_error (reader,
                                     "Invalid special type for integer read %d",
@@ -550,30 +548,19 @@ _bson_json_read_string (void                *_ctx, /* IN */
 #undef SSCANF
 
          break;
-      case BSON_JSON_LF_BINARY: {
-            /* TODO: error handling for pton */
+      case BSON_JSON_LF_BINARY:
+         {
             int binary_len;
             bson->bson_type_data.binary.has_binary = true;
             binary_len = b64_pton (val_w_null, NULL, 0);
+            if (binary_len < 0) {
+                goto BAD_PARSE;
+            }
             _bson_json_buf_ensure (&bson->bson_type_buf[0], binary_len + 1);
-            b64_pton ((char *)bson->bson_type_buf[2].buf,
-                      bson->bson_type_buf[0].buf, binary_len + 1);
+            b64_pton (val_w_null, bson->bson_type_buf[0].buf, binary_len + 1);
             bson->bson_type_buf[0].len = binary_len;
             break;
          }
-      case BSON_JSON_LF_REF:
-         bson->bson_type_data.ref.has_ref = true;
-         _bson_json_buf_set (&bson->bson_type_buf[0], val, vlen, true);
-         break;
-      case BSON_JSON_LF_ID:
-
-         if (vlen != 24) {
-            goto BAD_PARSE;
-         }
-
-         bson->bson_type_data.ref.has_id = true;
-         bson_oid_init_from_string (&bson->bson_type_data.ref.id, val_w_null);
-         break;
       case BSON_JSON_LF_INT64:
          {
             int64_t v64;
@@ -614,6 +601,20 @@ _bson_json_read_string (void                *_ctx, /* IN */
             }
          }
          break;
+      case BSON_JSON_LF_DECIMAL128:
+#ifdef BSON_EXPERIMENTAL_FEATURES
+         {
+            bson_decimal128_t decimal128;
+            bson_decimal128_from_string (val_w_null, &decimal128);
+
+            if (bson->read_state == BSON_JSON_IN_BSON_TYPE) {
+               bson->bson_type_data.v_decimal128.value = decimal128;
+            } else {
+               goto BAD_PARSE;
+            }
+         }
+         break;
+#endif /* BSON_EXPERIMENTAL_FEATURES */
       case BSON_JSON_LF_TIMESTAMP_T:
       case BSON_JSON_LF_TIMESTAMP_I:
       case BSON_JSON_LF_UNDEFINED:
@@ -674,13 +675,12 @@ _is_known_key (const char *key, size_t len)
           IS_KEY ("$binary") ||
           IS_KEY ("$type") ||
           IS_KEY ("$date") ||
-          IS_KEY ("$ref") ||
-          IS_KEY ("$id") ||
           IS_KEY ("$undefined") ||
           IS_KEY ("$maxKey") ||
           IS_KEY ("$minKey") ||
           IS_KEY ("$timestamp") ||
-          IS_KEY ("$numberLong"));
+          IS_KEY ("$numberLong")) ||
+          IS_KEY ("$numberDecimal");
 
 #undef IS_KEY
 
@@ -717,13 +717,14 @@ _bson_json_read_map_key (void          *_ctx, /* IN */
       if HANDLE_OPTION ("$binary", BSON_TYPE_BINARY, BSON_JSON_LF_BINARY) else
       if HANDLE_OPTION ("$type", BSON_TYPE_BINARY, BSON_JSON_LF_TYPE) else
       if HANDLE_OPTION ("$date", BSON_TYPE_DATE_TIME, BSON_JSON_LF_DATE) else
-      if HANDLE_OPTION ("$ref", BSON_TYPE_DBPOINTER, BSON_JSON_LF_REF) else
-      if HANDLE_OPTION ("$id", BSON_TYPE_DBPOINTER, BSON_JSON_LF_ID) else
       if HANDLE_OPTION ("$undefined", BSON_TYPE_UNDEFINED,
                         BSON_JSON_LF_UNDEFINED) else
       if HANDLE_OPTION ("$minKey", BSON_TYPE_MINKEY, BSON_JSON_LF_MINKEY) else
       if HANDLE_OPTION ("$maxKey", BSON_TYPE_MAXKEY, BSON_JSON_LF_MAXKEY) else
       if HANDLE_OPTION ("$numberLong", BSON_TYPE_INT64, BSON_JSON_LF_INT64) else
+#ifdef BSON_EXPERIMENTAL_FEATURES
+      if HANDLE_OPTION ("$numberDecimal", BSON_TYPE_DECIMAL128, BSON_JSON_LF_DECIMAL128) else
+#endif
       if (len == strlen ("$timestamp") &&
           memcmp (val, "$timestamp", len) == 0) {
          bson->bson_type = BSON_TYPE_TIMESTAMP;
@@ -844,28 +845,13 @@ _bson_json_read_append_timestamp (bson_json_reader_t      *reader, /* IN */
 }
 
 
-static int
-_bson_json_read_append_dbpointer (bson_json_reader_t      *reader, /* IN */
-                                  bson_json_reader_bson_t *bson)   /* IN */
+static void
+_bad_extended_json (bson_json_reader_t *reader)
 {
-   char *ref;
-
-   if (!bson->bson_type_data.ref.has_ref) {
-      _bson_json_read_set_error (reader,
-                                 "Missing $ref after $id in BSON_TYPE_DBPOINTER");
-      return 0;
-   }
-
-   if (!bson->bson_type_data.ref.has_id) {
-      _bson_json_read_set_error (reader,
-                                 "Missing $id after $ref in BSON_TYPE_DBPOINTER");
-      return 0;
-   }
-
-   ref = (char *)bson->bson_type_buf[0].buf;
-
-   return bson_append_dbpointer (STACK_BSON_CHILD, bson->key, (int)bson->key_buf.len,
-                                 ref, &bson->bson_type_data.ref.id);
+   bson_set_error (reader->error,
+                   BSON_ERROR_JSON,
+                   BSON_JSON_ERROR_READ_CORRUPT_JS,
+                   "Invalid MongoDB extended JSON");
 }
 
 
@@ -883,6 +869,12 @@ _bson_json_read_end_map (void *_ctx) /* IN */
    }
 
    if (bson->read_state == BSON_JSON_IN_BSON_TYPE) {
+      if (!bson->key) {
+         /* invalid, like {$numberLong: "1"} at the document top level */
+         _bad_extended_json (reader);
+         return false;
+      }
+
       bson->read_state = BSON_JSON_REGULAR;
       switch (bson->bson_type) {
       case BSON_TYPE_REGEX:
@@ -893,8 +885,6 @@ _bson_json_read_end_map (void *_ctx) /* IN */
          return _bson_json_read_append_binary (reader, bson);
       case BSON_TYPE_DATE_TIME:
          return _bson_json_read_append_date_time (reader, bson);
-      case BSON_TYPE_DBPOINTER:
-         return _bson_json_read_append_dbpointer (reader, bson);
       case BSON_TYPE_UNDEFINED:
          return bson_append_undefined (STACK_BSON_CHILD, bson->key,
                                        (int)bson->key_buf.len);
@@ -908,6 +898,12 @@ _bson_json_read_end_map (void *_ctx) /* IN */
          return bson_append_int64 (STACK_BSON_CHILD, bson->key,
                                    (int)bson->key_buf.len,
                                    bson->bson_type_data.v_int64.value);
+#ifdef BSON_EXPERIMENTAL_FEATURES
+      case BSON_TYPE_DECIMAL128:
+         return bson_append_decimal128 (STACK_BSON_CHILD, bson->key,
+                                    (int)bson->key_buf.len,
+                                    &bson->bson_type_data.v_decimal128.value);
+#endif /* BSON_EXPERIMENTAL_FEATURES */
       case BSON_TYPE_EOD:
       case BSON_TYPE_DOUBLE:
       case BSON_TYPE_UTF8:
@@ -920,18 +916,29 @@ _bson_json_read_end_map (void *_ctx) /* IN */
       case BSON_TYPE_CODEWSCOPE:
       case BSON_TYPE_INT32:
       case BSON_TYPE_TIMESTAMP:
+      case BSON_TYPE_DBPOINTER:
       default:
          _bson_json_read_set_error (reader, "Unknown type %d", bson->bson_type);
          return 0;
          break;
       }
    } else if (bson->read_state == BSON_JSON_IN_BSON_TYPE_TIMESTAMP_VALUES) {
+      if (!bson->key) {
+         _bad_extended_json (reader);
+         return false;
+      }
+
       bson->read_state = BSON_JSON_IN_BSON_TYPE_TIMESTAMP_ENDMAP;
 
       return _bson_json_read_append_timestamp (reader, bson);
    } else if (bson->read_state == BSON_JSON_IN_BSON_TYPE_TIMESTAMP_ENDMAP) {
       bson->read_state = BSON_JSON_REGULAR;
    } else if (bson->read_state == BSON_JSON_IN_BSON_TYPE_DATE_NUMBERLONG) {
+      if (!bson->key) {
+         _bad_extended_json (reader);
+         return false;
+      }
+
       bson->read_state = BSON_JSON_IN_BSON_TYPE_DATE_ENDMAP;
 
       return _bson_json_read_append_date_time(reader, bson);
@@ -1182,8 +1189,8 @@ bson_json_reader_new (void                 *data,           /* IN */
    p->data = data;
    p->cb = cb;
    p->dcb = dcb;
-   p->buf = bson_malloc (buf_size);
    p->buf_size = buf_size ? buf_size : BSON_JSON_DEFAULT_BUF_SIZE;
+   p->buf = bson_malloc (p->buf_size);
 
    r->yh = yajl_alloc (&read_cbs, &gYajlAllocFuncs, r);
 
