@@ -15,28 +15,38 @@
  */
 
 
+#include "mongoc.h"
+#include "mongoc-apm-private.h"
 #include "mongoc-counters-private.h"
 #include "mongoc-client-pool-private.h"
+#include "mongoc-client-pool.h"
+#include "mongoc-client-private.h"
 #include "mongoc-queue-private.h"
 #include "mongoc-thread-private.h"
-#include "mongoc-cluster-private.h"
-#include "mongoc-client-private.h"
+#include "mongoc-topology-private.h"
 #include "mongoc-trace.h"
 
+#ifdef MONGOC_ENABLE_SSL
+#include "mongoc-ssl-private.h"
+#endif
 
 struct _mongoc_client_pool_t
 {
-   mongoc_mutex_t    mutex;
-   mongoc_cond_t     cond;
-   mongoc_queue_t    queue;
-   mongoc_uri_t     *uri;
-   uint32_t          min_pool_size;
-   uint32_t          max_pool_size;
-   uint32_t          size;
+   mongoc_mutex_t          mutex;
+   mongoc_cond_t           cond;
+   mongoc_queue_t          queue;
+   mongoc_topology_t      *topology;
+   mongoc_uri_t           *uri;
+   uint32_t                min_pool_size;
+   uint32_t                max_pool_size;
+   uint32_t                size;
 #ifdef MONGOC_ENABLE_SSL
-   bool              ssl_opts_set;
-   mongoc_ssl_opt_t  ssl_opts;
+   bool                    ssl_opts_set;
+   mongoc_ssl_opt_t        ssl_opts;
 #endif
+   mongoc_apm_callbacks_t  apm_callbacks;
+   void                   *apm_context;
+   int32_t                 error_api_version;
 };
 
 
@@ -45,19 +55,22 @@ void
 mongoc_client_pool_set_ssl_opts (mongoc_client_pool_t   *pool,
                                  const mongoc_ssl_opt_t *opts)
 {
-   bson_return_if_fail (pool);
+   BSON_ASSERT (pool);
 
    mongoc_mutex_lock (&pool->mutex);
 
+   _mongoc_ssl_opts_cleanup (&pool->ssl_opts);
+
    memset (&pool->ssl_opts, 0, sizeof pool->ssl_opts);
    pool->ssl_opts_set = false;
 
    if (opts) {
-      memcpy (&pool->ssl_opts, opts, sizeof pool->ssl_opts);
+      _mongoc_ssl_opts_copy_to (opts, &pool->ssl_opts);
       pool->ssl_opts_set = true;
-
    }
 
+   mongoc_topology_scanner_set_ssl_opts (pool->topology->scanner, &pool->ssl_opts);
+
    mongoc_mutex_unlock (&pool->mutex);
 }
 #endif
@@ -66,15 +79,28 @@ mongoc_client_pool_set_ssl_opts (mongoc_client_pool_t   *pool,
 mongoc_client_pool_t *
 mongoc_client_pool_new (const mongoc_uri_t *uri)
 {
+   mongoc_topology_t *topology;
    mongoc_client_pool_t *pool;
    const bson_t *b;
    bson_iter_t iter;
+#ifdef MONGOC_EXPERIMENTAL_FEATURES
+   const char *appname;
+#endif
+
 
    ENTRY;
 
-   bson_return_val_if_fail(uri, NULL);
+   BSON_ASSERT (uri);
 
-   pool = bson_malloc0(sizeof *pool);
+#ifndef MONGOC_ENABLE_SSL
+   if (mongoc_uri_get_ssl (uri)) {
+      MONGOC_ERROR ("Can't create SSL client pool,"
+                    " SSL not enabled in this build.");
+      return NULL;
+   }
+#endif
+
+   pool = (mongoc_client_pool_t *)bson_malloc0(sizeof *pool);
    mongoc_mutex_init(&pool->mutex);
    _mongoc_queue_init(&pool->queue);
    pool->uri = mongoc_uri_copy(uri);
@@ -82,6 +108,10 @@ mongoc_client_pool_new (const mongoc_uri_t *uri)
    pool->max_pool_size = 100;
    pool->size = 0;
 
+   topology = mongoc_topology_new(uri, false);
+   pool->topology = topology;
+   pool->error_api_version = MONGOC_ERROR_API_VERSION_LEGACY;
+
    b = mongoc_uri_get_options(pool->uri);
 
    if (bson_iter_init_find_case(&iter, b, "minpoolsize")) {
@@ -96,6 +126,14 @@ mongoc_client_pool_new (const mongoc_uri_t *uri)
       }
    }
 
+#ifdef MONGOC_EXPERIMENTAL_FEATURES
+   appname = mongoc_uri_get_option_as_utf8 (pool->uri, "appname", NULL);
+   if (appname) {
+      /* the appname should have already been validated */
+      BSON_ASSERT (mongoc_client_pool_set_appname (pool, appname));
+   }
+#endif
+
    mongoc_counter_client_pools_active_inc();
 
    RETURN(pool);
@@ -109,15 +147,22 @@ mongoc_client_pool_destroy (mongoc_client_pool_t *pool)
 
    ENTRY;
 
-   bson_return_if_fail(pool);
+   BSON_ASSERT (pool);
 
-   while ((client = _mongoc_queue_pop_head(&pool->queue))) {
+   while ((client = (mongoc_client_t *)_mongoc_queue_pop_head(&pool->queue))) {
       mongoc_client_destroy(client);
    }
 
+   mongoc_topology_destroy (pool->topology);
+
    mongoc_uri_destroy(pool->uri);
    mongoc_mutex_destroy(&pool->mutex);
    mongoc_cond_destroy(&pool->cond);
+
+#ifdef MONGOC_ENABLE_SSL
+   _mongoc_ssl_opts_cleanup (&pool->ssl_opts);
+#endif
+
    bson_free(pool);
 
    mongoc_counter_client_pools_active_dec();
@@ -127,6 +172,20 @@ mongoc_client_pool_destroy (mongoc_client_pool_t *pool)
 }
 
 
+/*
+ * Start the background topology scanner.
+ *
+ * This function assumes the pool's mutex is locked
+ */
+static void
+_start_scanner_if_needed (mongoc_client_pool_t *pool)
+{
+   if (!_mongoc_topology_start_background_scanner (pool->topology)) {
+      MONGOC_ERROR ("Background scanner did not start!");
+      abort ();
+   }
+}
+
 mongoc_client_t *
 mongoc_client_pool_pop (mongoc_client_pool_t *pool)
 {
@@ -134,14 +193,18 @@ mongoc_client_pool_pop (mongoc_client_pool_t *pool)
 
    ENTRY;
 
-   bson_return_val_if_fail(pool, NULL);
+   BSON_ASSERT (pool);
 
    mongoc_mutex_lock(&pool->mutex);
 
 again:
-   if (!(client = _mongoc_queue_pop_head(&pool->queue))) {
+   if (!(client = (mongoc_client_t *)_mongoc_queue_pop_head(&pool->queue))) {
       if (pool->size < pool->max_pool_size) {
-         client = mongoc_client_new_from_uri(pool->uri);
+         client = _mongoc_client_new_from_uri(pool->uri, pool->topology);
+         client->error_api_version = pool->error_api_version;
+         _mongoc_client_set_apm_callbacks_private (client,
+                                                   &pool->apm_callbacks,
+                                                   pool->apm_context);
 #ifdef MONGOC_ENABLE_SSL
          if (pool->ssl_opts_set) {
             mongoc_client_set_ssl_opts (client, &pool->ssl_opts);
@@ -154,6 +217,7 @@ again:
       }
    }
 
+   _start_scanner_if_needed (pool);
    mongoc_mutex_unlock(&pool->mutex);
 
    RETURN(client);
@@ -167,13 +231,13 @@ mongoc_client_pool_try_pop (mongoc_client_pool_t *pool)
 
    ENTRY;
 
-   bson_return_val_if_fail(pool, NULL);
+   BSON_ASSERT (pool);
 
    mongoc_mutex_lock(&pool->mutex);
 
-   if (!(client = _mongoc_queue_pop_head(&pool->queue))) {
+   if (!(client = (mongoc_client_t *)_mongoc_queue_pop_head(&pool->queue))) {
       if (pool->size < pool->max_pool_size) {
-         client = mongoc_client_new_from_uri(pool->uri);
+         client = _mongoc_client_new_from_uri(pool->uri, pool->topology);
 #ifdef MONGOC_ENABLE_SSL
          if (pool->ssl_opts_set) {
             mongoc_client_set_ssl_opts (client, &pool->ssl_opts);
@@ -183,6 +247,9 @@ mongoc_client_pool_try_pop (mongoc_client_pool_t *pool)
       }
    }
 
+   if (client) {
+      _start_scanner_if_needed (pool);
+   }
    mongoc_mutex_unlock(&pool->mutex);
 
    RETURN(client);
@@ -195,32 +262,20 @@ mongoc_client_pool_push (mongoc_client_pool_t *pool,
 {
    ENTRY;
 
-   bson_return_if_fail(pool);
-   bson_return_if_fail(client);
+   BSON_ASSERT (pool);
+   BSON_ASSERT (client);
 
    mongoc_mutex_lock(&pool->mutex);
-   if (pool->size > pool->min_pool_size) {
+   if (pool->min_pool_size && pool->size > pool->min_pool_size) {
       mongoc_client_t *old_client;
-      old_client = _mongoc_queue_pop_head (&pool->queue);
+      old_client = (mongoc_client_t *)_mongoc_queue_pop_head (&pool->queue);
       if (old_client) {
           mongoc_client_destroy (old_client);
           pool->size--;
       }
    }
-   mongoc_mutex_unlock(&pool->mutex);
 
-   if ((client->cluster.state == MONGOC_CLUSTER_STATE_HEALTHY) ||
-       (client->cluster.state == MONGOC_CLUSTER_STATE_BORN)) {
-      mongoc_mutex_lock (&pool->mutex);
-      _mongoc_queue_push_tail (&pool->queue, client);
-   } else if (_mongoc_cluster_reconnect (&client->cluster, NULL)) {
-      mongoc_mutex_lock (&pool->mutex);
-      _mongoc_queue_push_tail (&pool->queue, client);
-   } else {
-      mongoc_client_destroy (client);
-      mongoc_mutex_lock (&pool->mutex);
-      pool->size--;
-   }
+   _mongoc_queue_push_head (&pool->queue, client);
 
    mongoc_cond_signal(&pool->cond);
    mongoc_mutex_unlock(&pool->mutex);
@@ -241,3 +296,83 @@ mongoc_client_pool_get_size (mongoc_client_pool_t *pool)
 
    RETURN (size);
 }
+
+void
+mongoc_client_pool_max_size(mongoc_client_pool_t *pool,
+                            uint32_t              max_pool_size)
+{
+   ENTRY;
+
+   mongoc_mutex_lock (&pool->mutex);
+   pool->max_pool_size = max_pool_size;
+   mongoc_mutex_unlock (&pool->mutex);
+
+   EXIT;
+}
+
+void
+mongoc_client_pool_min_size(mongoc_client_pool_t *pool,
+                            uint32_t              min_pool_size)
+{
+   ENTRY;
+
+   mongoc_mutex_lock (&pool->mutex);
+   pool->min_pool_size = min_pool_size;
+   mongoc_mutex_unlock (&pool->mutex);
+
+   EXIT;
+}
+
+bool
+mongoc_client_pool_set_apm_callbacks (mongoc_client_pool_t   *pool,
+                                      mongoc_apm_callbacks_t *callbacks,
+                                      void                   *context)
+{
+
+   if (pool->apm_callbacks.started ||
+       pool->apm_callbacks.succeeded ||
+       pool->apm_callbacks.failed ||
+       pool->apm_context) {
+      MONGOC_ERROR ("Can only set callbacks once");
+      return false;
+   }
+
+   if (callbacks) {
+      memcpy (&pool->apm_callbacks, callbacks, sizeof pool->apm_callbacks);
+   }
+
+   pool->apm_context = context;
+
+   return true;
+}
+
+bool
+mongoc_client_pool_set_error_api (mongoc_client_pool_t *pool,
+                                  int32_t               version)
+{
+   if (version != MONGOC_ERROR_API_VERSION_LEGACY &&
+       version != MONGOC_ERROR_API_VERSION_2) {
+      MONGOC_ERROR ("Unsupported Error API Version: %" PRId32, version);
+      return false;
+   }
+
+   pool->error_api_version = version;
+
+   return true;
+}
+
+#ifdef MONGOC_EXPERIMENTAL_FEATURES
+bool
+mongoc_client_pool_set_appname (mongoc_client_pool_t *pool,
+                                const char           *appname)
+{
+   bool ret;
+
+   mongoc_mutex_lock (&pool->mutex);
+   ret = _mongoc_topology_set_appname (pool->topology,
+                                       appname);
+   mongoc_mutex_unlock (&pool->mutex);
+
+   return ret;
+}
+#endif
