@@ -21,6 +21,8 @@
 #include "mongoc-bulk-operation.h"
 #include "mongoc-bulk-operation-private.h"
 #include "mongoc-client-private.h"
+#include "mongoc-find-and-modify-private.h"
+#include "mongoc-find-and-modify.h"
 #include "mongoc-collection.h"
 #include "mongoc-collection-private.h"
 #include "mongoc-cursor-private.h"
@@ -30,6 +32,7 @@
 #include "mongoc-index.h"
 #include "mongoc-log.h"
 #include "mongoc-trace.h"
+#include "mongoc-read-concern-private.h"
 #include "mongoc-write-concern-private.h"
 
 
@@ -54,34 +57,22 @@
          } \
       } while(0); \
 
-static bool
-validate_name (const char *str)
+
+static mongoc_cursor_t *
+_mongoc_collection_cursor_new (mongoc_collection_t *collection,
+                               mongoc_query_flags_t flags)
 {
-   const char *c;
-
-   if (str && *str) {
-      for (c = str; *c; c++) {
-         switch (*c) {
-         case '/':
-         case '\\':
-         case '.':
-         case '"':
-         case '*':
-         case '<':
-         case '>':
-         case ':':
-         case '|':
-         case '?':
-            return false;
-         default:
-            break;
-         }
-      }
-      return ((0 != strcmp (str, "oplog.$main")) &&
-              (0 != strcmp (str, "$cmd")));
-   }
-
-   return false;
+  return _mongoc_cursor_new (collection->client,
+                             collection->ns,
+                             flags,
+                             0,                  /* skip */
+                             0,                  /* limit */
+                             0,                  /* batch_size */
+                             false,              /* is_command */
+                             NULL,               /* query */
+                             NULL,               /* fields */
+                             NULL,               /* read prefs */
+                             NULL);              /* read concern */
 }
 
 static void
@@ -124,6 +115,7 @@ _mongoc_collection_write_command_execute (mongoc_write_command_t       *command,
  *       @db is the db name of the collection.
  *       @collection is the name of the collection.
  *       @read_prefs is the default read preferences to apply or NULL.
+ *       @read_concern is the default read concern to apply or NULL.
  *       @write_concern is the default write concern to apply or NULL.
  *
  * Returns:
@@ -141,6 +133,7 @@ _mongoc_collection_new (mongoc_client_t              *client,
                         const char                   *db,
                         const char                   *collection,
                         const mongoc_read_prefs_t    *read_prefs,
+                        const mongoc_read_concern_t  *read_concern,
                         const mongoc_write_concern_t *write_concern)
 {
    mongoc_collection_t *col;
@@ -156,6 +149,9 @@ _mongoc_collection_new (mongoc_client_t              *client,
    col->write_concern = write_concern ?
       mongoc_write_concern_copy(write_concern) :
       mongoc_write_concern_new();
+   col->read_concern = read_concern ?
+      mongoc_read_concern_copy(read_concern) :
+      mongoc_read_concern_new();
    col->read_prefs = read_prefs ?
       mongoc_read_prefs_copy(read_prefs) :
       mongoc_read_prefs_new(MONGOC_READ_PRIMARY);
@@ -208,6 +204,11 @@ mongoc_collection_destroy (mongoc_collection_t *collection) /* IN */
       collection->read_prefs = NULL;
    }
 
+   if (collection->read_concern) {
+      mongoc_read_concern_destroy(collection->read_concern);
+      collection->read_concern = NULL;
+   }
+
    if (collection->write_concern) {
       mongoc_write_concern_destroy(collection->write_concern);
       collection->write_concern = NULL;
@@ -219,6 +220,37 @@ mongoc_collection_destroy (mongoc_collection_t *collection) /* IN */
 }
 
 
+/*
+ *--------------------------------------------------------------------------
+ *
+ * mongoc_collection_copy --
+ *
+ *       Returns a copy of @collection that needs to be freed by calling
+ *       mongoc_collection_destroy.
+ *
+ * Returns:
+ *       A copy of this collection.
+ *
+ * Side effects:
+ *       None.
+ *
+ *--------------------------------------------------------------------------
+ */
+
+mongoc_collection_t *
+mongoc_collection_copy (mongoc_collection_t *collection) /* IN */
+{
+   ENTRY;
+
+   BSON_ASSERT (collection);
+
+   RETURN(_mongoc_collection_new(collection->client, collection->db,
+                                 collection->collection, collection->read_prefs,
+                                 collection->read_concern,
+                                 collection->write_concern));
+}
+
+
 /*
  *--------------------------------------------------------------------------
  *
@@ -270,8 +302,7 @@ mongoc_collection_aggregate (mongoc_collection_t       *collection, /* IN */
                              const bson_t              *options,    /* IN */
                              const mongoc_read_prefs_t *read_prefs) /* IN */
 {
-   mongoc_server_description_t *selected_server;
-   bson_error_t error;
+   mongoc_server_description_t *selected_server = NULL;
    mongoc_cursor_t *cursor;
    bson_iter_t iter;
    bson_t command;
@@ -279,26 +310,35 @@ mongoc_collection_aggregate (mongoc_collection_t       *collection, /* IN */
    int32_t batch_size = 0;
    bool use_cursor;
 
+   ENTRY;
+
    BSON_ASSERT (collection);
    BSON_ASSERT (pipeline);
 
+   bson_init (&command);
+
    if (!read_prefs) {
       read_prefs = collection->read_prefs;
    }
 
+   cursor = _mongoc_collection_cursor_new (collection, flags);
+
+   if (!_mongoc_read_prefs_validate (read_prefs, &cursor->error)) {
+      GOTO (done); 
+   }
+
    selected_server = mongoc_topology_select(collection->client->topology,
                                             MONGOC_SS_READ,
                                             read_prefs,
-                                            15,
-                                            &error);
+                                            &cursor->error);
 
    if (!selected_server) {
-      return NULL;
+      GOTO (done);
    }
 
-   use_cursor = selected_server->max_wire_version >= 1;
-
-   bson_init (&command);
+   cursor->server_id = selected_server->id;
+   use_cursor = 
+      selected_server->max_wire_version >= WIRE_VERSION_AGG_CURSOR;
 
    BSON_APPEND_UTF8 (&command, "aggregate", collection->collection);
 
@@ -309,12 +349,11 @@ mongoc_collection_aggregate (mongoc_collection_t       *collection, /* IN */
    if (bson_iter_init_find (&iter, pipeline, "pipeline") &&
        BSON_ITER_HOLDS_ARRAY (&iter)) {
       if (!bson_append_iter (&command, "pipeline", 8, &iter)) {
-         bson_set_error (&error,
+         bson_set_error (&cursor->error,
                MONGOC_ERROR_COMMAND,
                MONGOC_ERROR_COMMAND_INVALID_ARG,
                "Failed to append \"pipeline\" to create command.");
-         bson_destroy (&command);
-         return NULL;
+         GOTO (done);
       }
    } else {
       BSON_APPEND_ARRAY (&command, "pipeline", pipeline);
@@ -343,42 +382,49 @@ mongoc_collection_aggregate (mongoc_collection_t       *collection, /* IN */
       while (bson_iter_next (&iter)) {
          if (! (BSON_ITER_IS_KEY (&iter, "batchSize") || BSON_ITER_IS_KEY (&iter, "cursor"))) {
             if (!bson_append_iter (&command, bson_iter_key (&iter), -1, &iter)) {
-               bson_set_error (&error,
+               bson_set_error (&cursor->error,
                      MONGOC_ERROR_COMMAND,
                      MONGOC_ERROR_COMMAND_INVALID_ARG,
                      "Failed to append \"batchSize\" or \"cursor\" to create command.");
-               bson_destroy (&command);
-               return NULL;
+               GOTO (done);
             }
          }
       }
    }
 
-   cursor = mongoc_collection_command (collection, flags, 0, 0, batch_size,
-                                       &command, NULL, read_prefs);
+   if (collection->read_concern->level != NULL) {
+      const bson_t *read_concern_bson;
 
-   cursor->hint = selected_server->id;
+      if (selected_server->max_wire_version < WIRE_VERSION_READ_CONCERN) {
+         bson_set_error (&cursor->error,
+                         MONGOC_ERROR_COMMAND,
+                         MONGOC_ERROR_PROTOCOL_BAD_WIRE_VERSION,
+                         "The selected server does not support readConcern");
+         GOTO (done);
+      }
+
+      read_concern_bson = _mongoc_read_concern_get_bson (collection->read_concern);
+      BSON_APPEND_DOCUMENT (&command, "readConcern", read_concern_bson);
+   }
 
    if (use_cursor) {
-      /* even for newer versions, we get back a cursor document, that we have
-       * to patch in */
-      _mongoc_cursor_cursorid_init(cursor);
-      cursor->limit = 0;
-
-      /* we always return the cursor, even if it fails; users can detect the failure on performing
-       * a cursor operation. see CDRIVER-880. */
-      _mongoc_cursor_cursorid_prime (cursor);
+      _mongoc_cursor_cursorid_init (cursor, &command);
    } else {
       /* for older versions we get an array that we can create a synthetic
        * cursor on top of */
-      _mongoc_cursor_array_init(cursor, "result");
-      cursor->limit = 0;
+      _mongoc_cursor_array_init (cursor, &command, "result");
+   }
+
+done:
+   if (selected_server) {
+      mongoc_server_description_destroy (selected_server);
    }
 
    bson_destroy(&command);
-   mongoc_server_description_destroy (selected_server);
 
-   return cursor;
+   /* we always return the cursor, even if it fails; users can detect the
+    * failure on performing a cursor operation. see CDRIVER-880. */
+   RETURN (cursor);
 }
 
 
@@ -435,6 +481,7 @@ mongoc_collection_find (mongoc_collection_t       *collection, /* IN */
                         const bson_t              *fields,     /* IN */
                         const mongoc_read_prefs_t *read_prefs) /* IN */
 {
+   mongoc_cursor_t *cursor;
    BSON_ASSERT (collection);
    BSON_ASSERT (query);
 
@@ -443,9 +490,15 @@ mongoc_collection_find (mongoc_collection_t       *collection, /* IN */
    if (!read_prefs) {
       read_prefs = collection->read_prefs;
    }
-
-   return _mongoc_cursor_new(collection->client, collection->ns, flags, skip,
-                             limit, batch_size, false, query, fields, read_prefs);
+   
+   cursor = _mongoc_cursor_new (collection->client, collection->ns, flags, skip,
+                                limit, batch_size, false, query, fields, 
+                                read_prefs, collection->read_concern);
+   if (cursor->error.domain == 0) {
+      _mongoc_read_prefs_validate (read_prefs, &cursor->error);
+   }
+   
+   return cursor;   
 }
 
 
@@ -554,11 +607,14 @@ int64_t
 mongoc_collection_count (mongoc_collection_t       *collection,  /* IN */
                          mongoc_query_flags_t       flags,       /* IN */
                          const bson_t              *query,       /* IN */
-                         int64_t               skip,        /* IN */
-                         int64_t               limit,       /* IN */
+                         int64_t                    skip,        /* IN */
+                         int64_t                    limit,       /* IN */
                          const mongoc_read_prefs_t *read_prefs,  /* IN */
                          bson_error_t              *error)       /* OUT */
 {
+   /* Server Selection Spec: "may-use-secondary" commands SHOULD take a read
+    * preference argument and otherwise MUST use the default read preference
+    * from client, database or collection configuration. */
    return mongoc_collection_count_with_opts (
       collection,
       flags,
@@ -566,7 +622,7 @@ mongoc_collection_count (mongoc_collection_t       *collection,  /* IN */
       skip,
       limit,
       NULL,
-      read_prefs,
+      read_prefs ? read_prefs : collection->read_prefs,
       error);
 }
 
@@ -581,15 +637,30 @@ mongoc_collection_count_with_opts (mongoc_collection_t       *collection,  /* IN
                                    const mongoc_read_prefs_t *read_prefs,  /* IN */
                                    bson_error_t              *error)       /* OUT */
 {
-   int64_t ret = -1;
+   mongoc_server_stream_t *server_stream = NULL;
+   mongoc_cluster_t *cluster;
+   mongoc_apply_read_prefs_result_t read_prefs_result = READ_PREFS_RESULT_INIT;
    bson_iter_t iter;
+   int64_t ret = -1;
+   bool success;
    bson_t reply;
-   bson_t cmd;
+   bson_t cmd = BSON_INITIALIZER;
    bson_t q;
 
+   ENTRY;
+
+   cluster = &collection->client->cluster;
+   server_stream = mongoc_cluster_stream_for_writes (cluster, error);
+   if (!server_stream) {
+      GOTO (done);
+   }
+
+   if (!_mongoc_read_prefs_validate (read_prefs, error)) {
+      GOTO (done); 
+   }
+
    BSON_ASSERT (collection);
 
-   bson_init(&cmd);
    bson_append_utf8(&cmd, "count", 5, collection->collection,
                     collection->collectionlen);
    if (query) {
@@ -605,18 +676,43 @@ mongoc_collection_count_with_opts (mongoc_collection_t       *collection,  /* IN
    if (skip) {
       bson_append_int64(&cmd, "skip", 4, skip);
    }
+   if (collection->read_concern->level != NULL) {
+      const bson_t *read_concern_bson;
+
+      if (server_stream->sd->max_wire_version < WIRE_VERSION_READ_CONCERN) {
+         bson_set_error (error,
+                         MONGOC_ERROR_COMMAND,
+                         MONGOC_ERROR_PROTOCOL_BAD_WIRE_VERSION,
+                         "The selected server does not support readConcern");
+         GOTO (done);
+      }
+
+      read_concern_bson = _mongoc_read_concern_get_bson (collection->read_concern);
+      BSON_APPEND_DOCUMENT (&cmd, "readConcern", read_concern_bson);
+   }
    if (opts) {
        bson_concat(&cmd, opts);
    }
-   if (mongoc_collection_command_simple(collection, &cmd, read_prefs,
-                                        &reply, error) &&
-       bson_iter_init_find(&iter, &reply, "n")) {
+
+   apply_read_preferences (read_prefs, server_stream,
+                           &cmd, flags, &read_prefs_result);
+
+   success = mongoc_cluster_run_command_monitored (
+      cluster, server_stream, read_prefs_result.flags, collection->db,
+      read_prefs_result.query_with_read_prefs, &reply, error);
+
+   if (success && bson_iter_init_find(&iter, &reply, "n")) {
       ret = bson_iter_as_int64(&iter);
    }
-   bson_destroy(&reply);
-   bson_destroy(&cmd);
 
-   return ret;
+   bson_destroy (&reply);
+
+done:
+   apply_read_prefs_result_cleanup (&read_prefs_result);
+   mongoc_server_stream_cleanup (server_stream);
+   bson_destroy (&cmd);
+
+   RETURN (ret);
 }
 
 
@@ -1048,7 +1144,6 @@ mongoc_collection_find_indexes (mongoc_collection_t *collection,
                                 bson_error_t        *error)
 {
    mongoc_cursor_t *cursor;
-   mongoc_read_prefs_t *read_prefs;
    bson_t cmd = BSON_INITIALIZER;
    bson_t child;
 
@@ -1060,46 +1155,40 @@ mongoc_collection_find_indexes (mongoc_collection_t *collection,
    BSON_APPEND_DOCUMENT_BEGIN (&cmd, "cursor", &child);
    bson_append_document_end (&cmd, &child);
 
-   read_prefs = mongoc_read_prefs_new (MONGOC_READ_PRIMARY);
+   cursor = _mongoc_collection_cursor_new (collection, MONGOC_QUERY_SLAVE_OK);
+   _mongoc_cursor_cursorid_init (cursor, &cmd);
 
-   cursor = mongoc_collection_command (collection, MONGOC_QUERY_SLAVE_OK, 0, 0, 0,
-                                       &cmd, NULL, read_prefs);
-
-   _mongoc_cursor_cursorid_init(cursor);
-   cursor->limit = 0;
-
-   if (_mongoc_cursor_cursorid_prime(cursor)) {
+   if (_mongoc_cursor_cursorid_prime (cursor)) {
        /* intentionally empty */
    } else {
       if (mongoc_cursor_error (cursor, error)) {
+         mongoc_cursor_destroy (cursor);
+
          if (error->code == MONGOC_ERROR_COLLECTION_DOES_NOT_EXIST) {
-             bson_t empty_arr = BSON_INITIALIZER;
-             /* collection does not exist. in accordance with the spec we return
+            bson_t empty_arr = BSON_INITIALIZER;
+            /* collection does not exist. in accordance with the spec we return
              * an empty array. Also we need to clear out the error. */
-             error->code = 0;
-             error->domain = 0;
-             _mongoc_cursor_array_set_bson (cursor, &empty_arr);
+            error->code = 0;
+            error->domain = 0;
+            cursor = _mongoc_collection_cursor_new (collection,
+                                                    MONGOC_QUERY_SLAVE_OK);
+
+            _mongoc_cursor_array_init (cursor, NULL, NULL);
+            _mongoc_cursor_array_set_bson (cursor, &empty_arr);
          } else if (error->code == MONGOC_ERROR_QUERY_COMMAND_NOT_FOUND) {
             /* talking to an old server. */
             /* clear out error. */
             error->code = 0;
             error->domain = 0;
-            mongoc_cursor_destroy (cursor);
             cursor = _mongoc_collection_find_indexes_legacy (collection, error);
+         } else {
+            /* other error, to be handled by caller */
+            cursor = NULL;
          }
-      } else {
-         /* TODO: remove this branch for general release.  Only relevant for RC */
-         mongoc_cursor_destroy (cursor);
-         cursor = mongoc_collection_command (collection, MONGOC_QUERY_SLAVE_OK, 0, 0, 0,
-                                             &cmd, NULL, read_prefs);
-
-         _mongoc_cursor_array_init(cursor, "indexes");
-         cursor->limit = 0;
       }
    }
 
    bson_destroy (&cmd);
-   mongoc_read_prefs_destroy (read_prefs);
 
    return cursor;
 }
@@ -1144,8 +1233,8 @@ mongoc_collection_insert_bulk (mongoc_collection_t           *collection,
    mongoc_write_command_t command;
    mongoc_write_result_t result;
    mongoc_bulk_write_flags_t write_flags = MONGOC_BULK_WRITE_FLAGS_INIT;
-   bool ret;
    uint32_t i;
+   bool ret;
 
    BSON_ASSERT (collection);
    BSON_ASSERT (documents);
@@ -1176,7 +1265,9 @@ mongoc_collection_insert_bulk (mongoc_collection_t           *collection,
 
    write_flags.ordered = !(flags & MONGOC_INSERT_CONTINUE_ON_ERROR);
 
-   _mongoc_write_command_init_insert (&command, NULL, write_flags, true);
+   _mongoc_write_command_init_insert (&command, NULL, write_flags,
+                                      ++collection->client->cluster.operation_id,
+                                      true);
 
    for (i = 0; i < n_documents; i++) {
       _mongoc_write_command_insert_append (&command, documents[i]);
@@ -1186,7 +1277,11 @@ mongoc_collection_insert_bulk (mongoc_collection_t           *collection,
                                              write_concern, &result);
 
    collection->gle = bson_new ();
-   ret = _mongoc_write_result_complete (&result, collection->gle, error);
+   ret = _mongoc_write_result_complete (&result,
+                                        collection->client->error_api_version,
+                                        write_concern,
+                                        collection->gle,
+                                        error);
 
    _mongoc_write_result_destroy (&result);
    _mongoc_write_command_destroy (&command);
@@ -1261,13 +1356,19 @@ mongoc_collection_insert (mongoc_collection_t          *collection,
    }
 
    _mongoc_write_result_init (&result);
-   _mongoc_write_command_init_insert (&command, document, write_flags, false);
+   _mongoc_write_command_init_insert (&command, document, write_flags,
+                                      ++collection->client->cluster.operation_id,
+                                      false);
 
    _mongoc_collection_write_command_execute (&command, collection,
                                              write_concern, &result);
 
    collection->gle = bson_new ();
-   ret = _mongoc_write_result_complete (&result, collection->gle, error);
+   ret = _mongoc_write_result_complete (&result,
+                                        collection->client->error_api_version,
+                                        write_concern,
+                                        collection->gle,
+                                        error);
 
    _mongoc_write_result_destroy (&result);
    _mongoc_write_command_destroy (&command);
@@ -1352,13 +1453,18 @@ mongoc_collection_update (mongoc_collection_t          *collection,
                                       update,
                                       !!(flags & MONGOC_UPDATE_UPSERT),
                                       !!(flags & MONGOC_UPDATE_MULTI_UPDATE),
-                                      write_flags);
+                                      write_flags,
+                                      ++collection->client->cluster.operation_id);
 
    _mongoc_collection_write_command_execute (&command, collection,
                                              write_concern, &result);
 
    collection->gle = bson_new ();
-   ret = _mongoc_write_result_complete (&result, collection->gle, error);
+   ret = _mongoc_write_result_complete (&result,
+                                        collection->client->error_api_version,
+                                        write_concern,
+                                        collection->gle,
+                                        error);
 
    _mongoc_write_result_destroy (&result);
    _mongoc_write_command_destroy (&command);
@@ -1414,7 +1520,7 @@ mongoc_collection_save (mongoc_collection_t          *collection,
             MONGOC_ERROR_COMMAND_INVALID_ARG,
             "Failed to append bson to create update.");
       bson_destroy (&selector);
-      return NULL;
+      return false;
    }
 
    ret = mongoc_collection_update(collection,
@@ -1487,13 +1593,18 @@ mongoc_collection_remove (mongoc_collection_t          *collection,
    multi = !(flags & MONGOC_REMOVE_SINGLE_REMOVE);
 
    _mongoc_write_result_init (&result);
-   _mongoc_write_command_init_delete (&command, selector, multi, write_flags);
+   _mongoc_write_command_init_delete (&command, selector, multi, write_flags,
+                                      ++collection->client->cluster.operation_id);
 
    _mongoc_collection_write_command_execute (&command, collection,
                                              write_concern, &result);
 
    collection->gle = bson_new ();
-   ret = _mongoc_write_result_complete (&result, collection->gle, error);
+   ret = _mongoc_write_result_complete (&result,
+                                        collection->client->error_api_version,
+                                        write_concern,
+                                        collection->gle,
+                                        error);
 
    _mongoc_write_result_destroy (&result);
    _mongoc_write_command_destroy (&command);
@@ -1571,6 +1682,64 @@ mongoc_collection_set_read_prefs (mongoc_collection_t       *collection,
 }
 
 
+/*
+ *--------------------------------------------------------------------------
+ *
+ * mongoc_collection_get_read_concern --
+ *
+ *       Fetches the default read concern for the collection instance.
+ *
+ * Returns:
+ *       A mongoc_read_concern_t that should not be modified or freed.
+ *
+ * Side effects:
+ *       None.
+ *
+ *--------------------------------------------------------------------------
+ */
+
+const mongoc_read_concern_t *
+mongoc_collection_get_read_concern (const mongoc_collection_t *collection)
+{
+   BSON_ASSERT (collection);
+
+   return collection->read_concern;
+}
+
+
+/*
+ *--------------------------------------------------------------------------
+ *
+ * mongoc_collection_set_read_concern --
+ *
+ *       Sets the default read concern for the collection instance.
+ *
+ * Returns:
+ *       None.
+ *
+ * Side effects:
+ *       None.
+ *
+ *--------------------------------------------------------------------------
+ */
+
+void
+mongoc_collection_set_read_concern (mongoc_collection_t          *collection,
+                                    const mongoc_read_concern_t *read_concern)
+{
+   BSON_ASSERT (collection);
+
+   if (collection->read_concern) {
+      mongoc_read_concern_destroy (collection->read_concern);
+      collection->read_concern = NULL;
+   }
+
+   if (read_concern) {
+      collection->read_concern = mongoc_read_concern_copy (read_concern);
+   }
+}
+
+
 /*
  *--------------------------------------------------------------------------
  *
@@ -1719,7 +1888,8 @@ mongoc_collection_validate (mongoc_collection_t *collection, /* IN */
 {
    bson_iter_t iter;
    bson_t cmd = BSON_INITIALIZER;
-   bool ret;
+   bool ret = false;
+   bool reply_initialized = false;
 
    BSON_ASSERT (collection);
 
@@ -1730,7 +1900,7 @@ mongoc_collection_validate (mongoc_collection_t *collection, /* IN */
                       MONGOC_ERROR_BSON,
                       MONGOC_ERROR_BSON_INVALID,
                       "'full' must be a boolean value.");
-      return false;
+      goto cleanup;
    }
 
    bson_append_utf8 (&cmd, "validate", 8,
@@ -1742,9 +1912,15 @@ mongoc_collection_validate (mongoc_collection_t *collection, /* IN */
    }
 
    ret = mongoc_collection_command_simple (collection, &cmd, NULL, reply, error);
+   reply_initialized = true;
 
+cleanup:
    bson_destroy (&cmd);
 
+   if (reply && !reply_initialized) {
+      bson_init (reply);
+   }
+
    return ret;
 }
 
@@ -1785,7 +1961,7 @@ mongoc_collection_rename (mongoc_collection_t *collection,
    BSON_ASSERT (collection);
    BSON_ASSERT (new_name);
 
-   if (!validate_name (new_name)) {
+   if (strchr (new_name, '$')) {
       bson_set_error (error,
                       MONGOC_ERROR_NAMESPACE,
                       MONGOC_ERROR_NAMESPACE_INVALID,
@@ -1878,7 +2054,12 @@ mongoc_collection_stats (mongoc_collection_t *collection,
       bson_concat (&cmd, options);
    }
 
-   ret = mongoc_collection_command_simple (collection, &cmd, NULL, stats, error);
+   /* Server Selection Spec: "may-use-secondary" commands SHOULD take a read
+    * preference argument and otherwise MUST use the default read preference
+    * from client, database or collection configuration. */
+   ret = mongoc_collection_command_simple (collection, &cmd,
+                                           collection->read_prefs,
+                                           stats, error);
 
    bson_destroy (&cmd);
 
@@ -1904,11 +2085,156 @@ mongoc_collection_create_bulk_operation (
    return _mongoc_bulk_operation_new (collection->client,
                                       collection->db,
                                       collection->collection,
-                                      0,
                                       write_flags,
                                       write_concern);
 }
 
+/*
+ *--------------------------------------------------------------------------
+ *
+ * mongoc_collection_find_and_modify_with_opts --
+ *
+ *       Find a document in @collection matching @query, applying @opts.
+ *
+ *       If @reply is not NULL, then the result document will be placed
+ *       in reply and should be released with bson_destroy().
+ *
+ *       See http://docs.mongodb.org/manual/reference/command/findAndModify/
+ *       for more information.
+ *
+ * Returns:
+ *       true on success; false on failure.
+ *
+ * Side effects:
+ *       reply is initialized.
+ *       error is set if false is returned.
+ *
+ *--------------------------------------------------------------------------
+ */
+bool
+mongoc_collection_find_and_modify_with_opts (mongoc_collection_t                 *collection,
+                                             const bson_t                        *query,
+                                             const mongoc_find_and_modify_opts_t *opts,
+                                             bson_t                              *reply,
+                                             bson_error_t                        *error)
+{
+   mongoc_cluster_t *cluster;
+   mongoc_server_stream_t *server_stream;
+   bson_iter_t iter;
+   bson_iter_t inner;
+   const char *name;
+   bson_t reply_local;
+   bson_t *reply_ptr;
+   bool ret;
+   bson_t command = BSON_INITIALIZER;
+
+   ENTRY;
+
+   BSON_ASSERT (collection);
+   BSON_ASSERT (query);
+
+   reply_ptr = reply ? reply : &reply_local;
+   bson_init (reply_ptr);
+   cluster = &collection->client->cluster;
+   server_stream = mongoc_cluster_stream_for_writes (cluster, error);
+   if (!server_stream) {
+      bson_destroy (&command);
+      RETURN (false);
+   }
+
+   name = mongoc_collection_get_name (collection);
+   BSON_APPEND_UTF8 (&command, "findAndModify", name);
+   BSON_APPEND_DOCUMENT (&command, "query", query);
+
+   if (opts->sort) {
+      BSON_APPEND_DOCUMENT (&command, "sort", opts->sort);
+   }
+
+   if (opts->update) {
+      BSON_APPEND_DOCUMENT (&command, "update", opts->update);
+   }
+
+   if (opts->fields) {
+      BSON_APPEND_DOCUMENT (&command, "fields", opts->fields);
+   }
+
+   if (opts->flags & MONGOC_FIND_AND_MODIFY_REMOVE) {
+      BSON_APPEND_BOOL (&command, "remove", true);
+   }
+
+   if (opts->flags & MONGOC_FIND_AND_MODIFY_UPSERT) {
+      BSON_APPEND_BOOL (&command, "upsert", true);
+   }
+
+   if (opts->flags & MONGOC_FIND_AND_MODIFY_RETURN_NEW) {
+      BSON_APPEND_BOOL (&command, "new", true);
+   }
+
+   if (opts->bypass_document_validation != MONGOC_BYPASS_DOCUMENT_VALIDATION_DEFAULT) {
+      BSON_APPEND_BOOL (&command, "bypassDocumentValidation",
+                        !!opts->bypass_document_validation);
+   }
+
+   if (opts->max_time_ms > 0) {
+      BSON_APPEND_INT32 (&command, "maxTimeMS", opts->max_time_ms);
+   }
+
+   if (server_stream->sd->max_wire_version >= WIRE_VERSION_FAM_WRITE_CONCERN) {
+      if (!mongoc_write_concern_is_valid (collection->write_concern)) {
+         bson_set_error (error,
+                         MONGOC_ERROR_COMMAND,
+                         MONGOC_ERROR_COMMAND_INVALID_ARG,
+                         "The write concern is invalid.");
+         bson_destroy (&command);
+         mongoc_server_stream_cleanup (server_stream);
+         RETURN (false);
+      }
+
+      if (mongoc_write_concern_is_acknowledged (collection->write_concern)) {
+         _BSON_APPEND_WRITE_CONCERN (&command, collection->write_concern);
+      }
+   }
+
+   if (!bson_concat (&command, &opts->extra)) {
+      bson_set_error (error,
+                      MONGOC_ERROR_COMMAND,
+                      MONGOC_ERROR_COMMAND_INVALID_ARG,
+                      "mongoc_find_and_modify_opts_t.extra is corrupt.");
+      bson_destroy (&command);
+      mongoc_server_stream_cleanup (server_stream);
+      RETURN (false);
+   }
+
+   ret = mongoc_cluster_run_command_monitored (cluster, server_stream,
+                                               MONGOC_QUERY_NONE,
+                                               collection->db, &command,
+                                               reply_ptr, error);
+
+   if (bson_iter_init_find (&iter, reply_ptr, "writeConcernError") &&
+         BSON_ITER_HOLDS_DOCUMENT (&iter)) {
+      const char *errmsg = NULL;
+      int32_t code = 0;
+
+      bson_iter_recurse(&iter, &inner);
+      while (bson_iter_next (&inner)) {
+         if (BSON_ITER_IS_KEY (&inner, "code")) {
+            code = bson_iter_int32 (&inner);
+         } else if (BSON_ITER_IS_KEY (&inner, "errmsg")) {
+            errmsg = bson_iter_utf8 (&inner, NULL);
+         }
+      }
+      bson_set_error (error, MONGOC_ERROR_WRITE_CONCERN, code, "Write Concern error: %s", errmsg);
+      ret = false; 
+   }
+   if (reply_ptr == &reply_local) {
+      bson_destroy (reply_ptr);
+   }
+
+   bson_destroy (&command);
+   mongoc_server_stream_cleanup (server_stream);
+
+   RETURN (ret);
+}
 
 /*
  *--------------------------------------------------------------------------
@@ -1954,11 +2280,9 @@ mongoc_collection_find_and_modify (mongoc_collection_t *collection,
                                    bson_t              *reply,
                                    bson_error_t        *error)
 {
-   mongoc_cluster_t *cluster;
-   mongoc_server_stream_t *server_stream;
-   const char *name;
+   mongoc_find_and_modify_opts_t *opts;
+   int flags = 0;
    bool ret;
-   bson_t command = BSON_INITIALIZER;
 
    ENTRY;
 
@@ -1966,70 +2290,26 @@ mongoc_collection_find_and_modify (mongoc_collection_t *collection,
    BSON_ASSERT (query);
    BSON_ASSERT (update || _remove);
 
-   cluster = &collection->client->cluster;
-   server_stream = mongoc_cluster_stream_for_writes (cluster, error);
-   if (!server_stream) {
-      bson_destroy (&command);
-      RETURN (false);
-   }
-
-   name = mongoc_collection_get_name (collection);
-   BSON_APPEND_UTF8 (&command, "findAndModify", name);
-   BSON_APPEND_DOCUMENT (&command, "query", query);
-
-   if (sort) {
-      BSON_APPEND_DOCUMENT (&command, "sort", sort);
-   }
-
-   if (update) {
-      BSON_APPEND_DOCUMENT (&command, "update", update);
-   }
-
-   if (fields) {
-      BSON_APPEND_DOCUMENT (&command, "fields", fields);
-   }
 
    if (_remove) {
-      BSON_APPEND_BOOL (&command, "remove", _remove);
+      flags |= MONGOC_FIND_AND_MODIFY_REMOVE;
    }
-
    if (upsert) {
-      BSON_APPEND_BOOL (&command, "upsert", upsert);
+      flags |= MONGOC_FIND_AND_MODIFY_UPSERT;
    }
-
    if (_new) {
-      BSON_APPEND_BOOL (&command, "new", _new);
+      flags |= MONGOC_FIND_AND_MODIFY_RETURN_NEW;
    }
 
-   if (server_stream->sd->max_wire_version >= 4) {
-      if (!_mongoc_write_concern_is_valid (collection->write_concern)) {
-         bson_set_error (error,
-                         MONGOC_ERROR_COMMAND,
-                         MONGOC_ERROR_COMMAND_INVALID_ARG,
-                         "The write concern is invalid.");
-         bson_destroy (&command);
-         mongoc_server_stream_cleanup (server_stream);
-         RETURN (false);
-      }
+   opts = mongoc_find_and_modify_opts_new ();
 
-      if (_mongoc_write_concern_needs_gle (collection->write_concern)) {
-         _BSON_APPEND_WRITE_CONCERN (&command, collection->write_concern);
-      }
-   }
+   mongoc_find_and_modify_opts_set_sort (opts, sort);
+   mongoc_find_and_modify_opts_set_update (opts, update);
+   mongoc_find_and_modify_opts_set_fields (opts, fields);
+   mongoc_find_and_modify_opts_set_flags (opts, flags);
 
-   /*
-    * Submit the command to MongoDB server.
-    */
+   ret = mongoc_collection_find_and_modify_with_opts (collection, query, opts, reply, error);
+   mongoc_find_and_modify_opts_destroy (opts);
 
-   ret = mongoc_cluster_run_command (cluster, server_stream->stream,
-                                     MONGOC_QUERY_NONE, collection->db,
-                                     &command, reply, error);
-
-   /*
-    * Cleanup.
-    */
-   bson_destroy (&command);
-   mongoc_server_stream_cleanup (server_stream);
-
-   RETURN (ret);
+   return ret;
 }
