@@ -32,13 +32,11 @@
 
 typedef struct
 {
-   const bson_t       *result;
-   bool         has_array;
-   bool         has_synthetic_bson;
-   bson_iter_t         iter;
-   bson_t              bson;
-   uint32_t       document_len;
-   const uint8_t *document;
+   bson_t         array;
+   bool           has_array;
+   bool           has_synthetic_bson;
+   bson_iter_t    iter;
+   bson_t         bson;   /* current document */
    const char    *field_name;
 } mongoc_cursor_array_t;
 
@@ -51,6 +49,8 @@ _mongoc_cursor_array_new (const char *field_name)
    ENTRY;
 
    arr = (mongoc_cursor_array_t *)bson_malloc0 (sizeof *arr);
+   arr->has_array = false;
+   arr->has_synthetic_bson = false;
    arr->field_name = field_name;
 
    RETURN (arr);
@@ -66,6 +66,10 @@ _mongoc_cursor_array_destroy (mongoc_cursor_t *cursor)
 
    arr = (mongoc_cursor_array_t *)cursor->iface_data;
 
+   if (arr->has_array) {
+      bson_destroy (&arr->array);
+   }
+
    if (arr->has_synthetic_bson) {
       bson_destroy(&arr->bson);
    }
@@ -80,27 +84,24 @@ _mongoc_cursor_array_destroy (mongoc_cursor_t *cursor)
 bool
 _mongoc_cursor_array_prime (mongoc_cursor_t *cursor)
 {
-   bool ret = true;
    mongoc_cursor_array_t *arr;
    bson_iter_t iter;
 
    ENTRY;
 
    arr = (mongoc_cursor_array_t *)cursor->iface_data;
-   if (!arr->has_array) {
+
+   BSON_ASSERT (arr);
+
+   if (_mongoc_cursor_run_command (cursor, &cursor->query, &arr->array) &&
+       bson_iter_init_find (&iter, &arr->array, arr->field_name) &&
+       BSON_ITER_HOLDS_ARRAY (&iter) &&
+       bson_iter_recurse (&iter, &arr->iter)) {
       arr->has_array = true;
-
-      ret = _mongoc_cursor_next (cursor, &arr->result);
-
-      if (!(ret &&
-            bson_iter_init_find (&iter, arr->result, arr->field_name) &&
-            BSON_ITER_HOLDS_ARRAY (&iter) &&
-            bson_iter_recurse (&iter, &arr->iter))) {
-         ret = false;
-      }
+      return true;
    }
 
-   return ret;
+   return false;
 }
 
 
@@ -110,13 +111,15 @@ _mongoc_cursor_array_next (mongoc_cursor_t *cursor,
 {
    bool ret = true;
    mongoc_cursor_array_t *arr;
+   uint32_t document_len;
+   const uint8_t *document;
 
    ENTRY;
 
    arr = (mongoc_cursor_array_t *)cursor->iface_data;
    *bson = NULL;
 
-   if (!arr->has_array) {
+   if (!arr->has_array && !arr->has_synthetic_bson) {
       ret = _mongoc_cursor_array_prime(cursor);
    }
 
@@ -125,9 +128,8 @@ _mongoc_cursor_array_next (mongoc_cursor_t *cursor,
    }
 
    if (ret) {
-      bson_iter_document (&arr->iter, &arr->document_len, &arr->document);
-      bson_init_static (&arr->bson, arr->document, arr->document_len);
-
+      bson_iter_document (&arr->iter, &document_len, &document);
+      bson_init_static (&arr->bson, document, document_len);
       *bson = &arr->bson;
    }
 
@@ -146,7 +148,7 @@ _mongoc_cursor_array_clone (const mongoc_cursor_t *cursor)
    arr = (mongoc_cursor_array_t *)cursor->iface_data;
 
    clone_ = _mongoc_cursor_clone (cursor);
-   _mongoc_cursor_array_init (clone_, arr->field_name);
+   _mongoc_cursor_array_init (clone_, &cursor->query, arr->field_name);
 
    RETURN (clone_);
 }
@@ -163,7 +165,7 @@ _mongoc_cursor_array_more (mongoc_cursor_t *cursor)
 
    arr = (mongoc_cursor_array_t *)cursor->iface_data;
 
-   if (arr->has_array) {
+   if (arr->has_array || arr->has_synthetic_bson) {
       memcpy (&iter, &arr->iter, sizeof iter);
 
       ret = bson_iter_next (&iter);
@@ -204,10 +206,17 @@ static mongoc_cursor_interface_t gMongocCursorArray = {
 
 void
 _mongoc_cursor_array_init (mongoc_cursor_t *cursor,
+                           const bson_t    *command,
                            const char      *field_name)
 {
    ENTRY;
 
+
+   if (command) {
+      bson_destroy (&cursor->query);
+      bson_copy_to (command, &cursor->query);
+   }
+
    cursor->iface_data = _mongoc_cursor_array_new (field_name);
 
    memcpy (&cursor->iface, &gMongocCursorArray,
@@ -226,11 +235,9 @@ _mongoc_cursor_array_set_bson (mongoc_cursor_t *cursor,
    ENTRY;
 
    arr = (mongoc_cursor_array_t *)cursor->iface_data;
-
    bson_copy_to(bson, &arr->bson);
-
-   arr->has_array = true;
    arr->has_synthetic_bson = true;
-
    bson_iter_init(&arr->iter, &arr->bson);
+
+   EXIT;
 }
