@@ -22,21 +22,21 @@
 
 #include "mongoc-error.h"
 #include "mongoc-sasl-private.h"
+#include "mongoc-util-private.h"
 
 
 #ifndef SASL_CALLBACK_FN
 #  define SASL_CALLBACK_FN(_f) ((int (*) (void))(_f))
 #endif
 
-
 void
 _mongoc_sasl_set_mechanism (mongoc_sasl_t *sasl,
                             const char    *mechanism)
 {
    BSON_ASSERT (sasl);
 
-   free (sasl->mechanism);
-   sasl->mechanism = mechanism ? strdup (mechanism) : NULL;
+   bson_free (sasl->mechanism);
+   sasl->mechanism = mechanism ? bson_strdup (mechanism) : NULL;
 }
 
 
@@ -67,8 +67,8 @@ _mongoc_sasl_set_pass (mongoc_sasl_t *sasl,
 {
    BSON_ASSERT (sasl);
 
-   free (sasl->pass);
-   sasl->pass = pass ? strdup (pass) : NULL;
+   bson_free (sasl->pass);
+   sasl->pass = pass ? bson_strdup (pass) : NULL;
 }
 
 
@@ -99,8 +99,8 @@ _mongoc_sasl_set_user (mongoc_sasl_t *sasl,
 {
    BSON_ASSERT (sasl);
 
-   free (sasl->user);
-   sasl->user = user ? strdup (user) : NULL;
+   bson_free (sasl->user);
+   sasl->user = user ? bson_strdup (user) : NULL;
 }
 
 
@@ -110,8 +110,8 @@ _mongoc_sasl_set_service_host (mongoc_sasl_t *sasl,
 {
    BSON_ASSERT (sasl);
 
-   free (sasl->service_host);
-   sasl->service_host = service_host ? strdup (service_host) : NULL;
+   bson_free (sasl->service_host);
+   sasl->service_host = service_host ? bson_strdup (service_host) : NULL;
 }
 
 
@@ -121,8 +121,67 @@ _mongoc_sasl_set_service_name (mongoc_sasl_t *sasl,
 {
    BSON_ASSERT (sasl);
 
-   free (sasl->service_name);
-   sasl->service_name = service_name ? strdup (service_name) : NULL;
+   bson_free (sasl->service_name);
+   sasl->service_name = service_name ? bson_strdup (service_name) : NULL;
+}
+
+
+void
+_mongoc_sasl_set_properties (mongoc_sasl_t      *sasl,
+                             const mongoc_uri_t *uri)
+{
+   const bson_t *options;
+   bson_iter_t iter;
+   bson_t properties;
+   const char *service_name = NULL;
+   bool canonicalize = false;
+
+   options = mongoc_uri_get_options (uri);
+
+   if (!mongoc_uri_get_mechanism_properties (uri, &properties)) {
+      bson_init (&properties);
+   }
+
+   if (bson_iter_init_find_case (&iter, options, "gssapiservicename") &&
+       BSON_ITER_HOLDS_UTF8 (&iter)) {
+      service_name = bson_iter_utf8 (&iter, NULL);
+   }
+
+   if (bson_iter_init_find_case (&iter, &properties, "SERVICE_NAME") &&
+       BSON_ITER_HOLDS_UTF8 (&iter)) {
+      /* newer "authMechanismProperties" URI syntax takes precedence */
+      service_name = bson_iter_utf8 (&iter, NULL);
+   }
+
+   _mongoc_sasl_set_service_name (sasl, service_name);
+
+   /*
+    * Driver Authentication Spec: "Drivers MAY allow the user to request
+    * canonicalization of the hostname. This might be required when the hosts
+    * report different hostnames than what is used in the kerberos database.
+    * The default is "false".
+    *
+    * Some underlying GSSAPI layers will do this for us, but can be disabled in
+    * their config (krb.conf).
+    *
+    * See CDRIVER-323 for more information.
+    */
+   if (bson_iter_init_find_case (&iter, options, "canonicalizeHostname") &&
+       BSON_ITER_HOLDS_BOOL (&iter)) {
+      canonicalize = bson_iter_bool (&iter);
+   }
+
+   if (bson_iter_init_find_case (&iter,
+                                 &properties,
+                                 "CANONICALIZE_HOST_NAME") &&
+       BSON_ITER_HOLDS_UTF8 (&iter)) {
+      /* newer "authMechanismProperties" URI syntax takes precedence */
+      canonicalize = !strcasecmp (bson_iter_utf8 (&iter, NULL), "true");
+   }
+
+   sasl->canonicalize_host_name = canonicalize;
+
+   bson_destroy (&properties);
 }
 
 
@@ -163,11 +222,11 @@ _mongoc_sasl_destroy (mongoc_sasl_t *sasl)
       sasl_dispose (&sasl->conn);
    }
 
-   free (sasl->user);
-   free (sasl->pass);
-   free (sasl->mechanism);
-   free (sasl->service_name);
-   free (sasl->service_host);
+   bson_free (sasl->user);
+   bson_free (sasl->pass);
+   bson_free (sasl->mechanism);
+   bson_free (sasl->service_name);
+   bson_free (sasl->service_host);
 }
 
 
@@ -183,7 +242,7 @@ _mongoc_sasl_is_failure (int           status,
          bson_set_error (error,
                          MONGOC_ERROR_SASL,
                          status,
-                         "SASL Failrue: insufficient memory.");
+                         "SASL Failure: insufficient memory.");
          break;
       case SASL_NOMECH:
          bson_set_error (error,
